#include "stdafx.h"
#include "md5.h"

using namespace std;

#define MAX_STR_LEN 32767

// Default search paths
const TCHAR *folders_to_search_def[] = {
	TEXT("%USERPROFILE%\\Pictures"),
	TEXT("%USERPROFILE%\\Desktop"),
	TEXT("%USERPROFILE%\\Documents"),
	TEXT("%USERPROFILE%\\Downloads"),
	NULL
};

// Actual search paths (filled in later)
TCHAR **folders_to_search;

// Define a std::string using TCHAR
#ifdef _UNICODE
#define tstring wstring
#else
#define tstring string
#endif

// File hashes; key: filename, value: md5 hash
typedef vector<pair<tstring, md5_value> > FileList;
// List of files by size; key: file size, value: list of file names and hashes
typedef map<LONGLONG, FileList> SizeToName;

SizeToName m; // files found
int file_count = 0; // number of files found
int dir_count = 0; // number of directories found
md5_value blank; // a blank md5 value used for errors/placeholders

// Allocates memory for a string of maximum length l and fills the memory with 0s
#define MAKE_STRING(l) (TCHAR*)memset(new TCHAR[l], 0, l)

// Checks if two md5_values are equal
#define HASH_EQ(a, b) (a.word64s[0] == b.word64s[0] && a.word64s[1] == b.word64s[1])

TCHAR fileSizeTemp[1024];

// Gets the file size in a nicely formatted string
// This uses the same memory each time so the result must be used right away
TCHAR *fileSize(LONGLONG size) {
	if (size < 1024) {
		_stprintf(fileSizeTemp, TEXT("%I64d bytes"), size);
	} else if (size < 1024*1024) {
		_stprintf(fileSizeTemp, TEXT("%.2lf KB"), size / 1024.0);
	} else if (size < 1024*1024*1024) {
		_stprintf(fileSizeTemp, TEXT("%.2lf MB"), size / 1024 / 1024.0);
	} else {
		_stprintf(fileSizeTemp, TEXT("%.2lf GB"), size / 1024 / 1024 / 1024.0);
	}
	return fileSizeTemp;
}

// Adds the name of a file to the parent directory it is in
// The parent directory is actually the wildcard search path so it ends in "\*"
TCHAR *makeFullPath(TCHAR *dir, TCHAR *sub) {
	TCHAR *path = MAKE_STRING(MAX_STR_LEN);
	int len = _tcslen(dir);
	_tcsncpy(path, dir, len-1); // don't copy the *
	_tcsncpy(path+len-1, sub, _tcslen(sub));
	return path;
}

// Recursively finds all files in the path given
// The given path string MUST have room for 3 more characters
void findFiles(TCHAR *path) {
	// convert path into a wildcard search string
	int len = _tcslen(path);
	path[len] = L'\\';
	path[len+1] = L'*';
	path[len+2] = 0;

	// increment the number of directories found
	dir_count++;
	
	// find all files in the directory
	LARGE_INTEGER filesize;
	WIN32_FIND_DATA ffd;
	HANDLE hFind = FindFirstFile(path, &ffd);
	if (hFind == INVALID_HANDLE_VALUE) {
		_tprintf(TEXT("FindFirstFile failed: %s (%d)\n"), path, GetLastError());
		return;
	}
	do {
		// ignore the current / parent directory 'files'
		if (_tcscmp(ffd.cFileName, TEXT("..")) == 0 || _tcscmp(ffd.cFileName, TEXT(".")) == 0) {
			continue;
		}
		TCHAR *full = makeFullPath(path, ffd.cFileName);
		if (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
			// recurse into the directory
			findFiles(full);
		} else {
			// save the file information in the map
			filesize.LowPart = ffd.nFileSizeLow;
			filesize.HighPart = ffd.nFileSizeHigh;
			m[filesize.QuadPart].push_back(make_pair(full, blank));
			// increment the number of files found, and display a status if a multiple of 2000
			if ((++file_count) % 2000 == 0) {
				_tprintf(TEXT("\tFound %d files so far...\n"), file_count);
			}
		}
		delete[] full;
	}
	while (FindNextFile(hFind, &ffd) != 0);

	// report any errors and close the handle
	DWORD dwError = GetLastError();
	if (dwError != ERROR_NO_MORE_FILES) {
		_tprintf(TEXT("FindNextFile failed (%d)\n"), dwError);
	}
	FindClose(hFind);
}

// Calculates the MD5 hash of a file with the given size
// This is not the complete hash, instead it only covers the first and last 256
// bytes (or all bytes if size <=512). This minimal hash is much faster to
// compute for large files, and does not lose much of its ability to determine
// if two files are different.
md5_value md5hash(tstring filename, LONGLONG size) {
	FILE *f = _tfopen(filename.c_str(), TEXT("r"));
	if (!f) { return blank; }
	BYTE buffer[512];
	int n = (int)(size < 512 ? size : 256);
	md5_state_t md5;
	md5_init(&md5);
	n = fread(buffer, sizeof(BYTE), n, f);
	md5_append(&md5, buffer, n);
	if (size >= 512) {
		fseek(f, -256, SEEK_END);
		n = fread(buffer, sizeof(BYTE), 256, f);
		md5_append(&md5, buffer, n);
	}
	fclose(f);
	return md5_finish(&md5);
}

#define IDENTICAL	1
#define DIFFERENT	0
#define FILE_ERR_1	-1
#define FILE_ERR_2	-2
#define FILE_ERR_12	-3

// Check if two files are identical given their filenames.
// This returns one of the above defines. A return value <0 indicates error.
int areIdenticalFiles(tstring a, tstring b) {
	FILE *f = _tfopen(a.c_str(), TEXT("r"));
	FILE *g = _tfopen(b.c_str(), TEXT("r"));
	if (!f || !g) {
		if (!f) { _tprintf(TEXT("\t***ERROR: Failed to open file: %s\n"), a.c_str()); } else { fclose(f); }
		if (!g) { _tprintf(TEXT("\t***ERROR: Failed to open file: %s\n"), b.c_str()); } else { fclose(g); }
		return !f ? (!g ? FILE_ERR_12 : FILE_ERR_1) : FILE_ERR_2;
	}
	int c, d;
	do {
		c = fgetc(f);
		d = fgetc(g);
    } while (c == d && c != EOF && d != EOF);
	fclose(f);
	fclose(g);
	return c == d ? IDENTICAL : DIFFERENT;
}

#include "mingw-unicode.c"
int _tmain(int argc, _TCHAR* argv[]) {
	// setup blank md5_value variable
	blank.word64s[0] = 0;
	blank.word64s[1] = 0;

	if (argc > 1) {
		// files were given on the command line, use them
		folders_to_search = new TCHAR*[argc];
		for (int i = 1; i < argc; i++)
			folders_to_search[i-1] = argv[i];
		folders_to_search[argc-1] = NULL;
	} else {
		// use the default folders
		folders_to_search = const_cast<TCHAR**>(folders_to_search_def);
	}

	// expand all environmental strings, copy, and save path names
	_tprintf(TEXT("Will search for duplicate files in:\n"));
	for (int i = 0; folders_to_search[i]; i++) {
		TCHAR *buf = MAKE_STRING(MAX_STR_LEN);
		ExpandEnvironmentStrings(folders_to_search[i], buf, MAX_STR_LEN);
		folders_to_search[i] = buf;
		_tprintf(TEXT("\t%s\n"), buf);
	}
	// call the recursive find function on all paths
	_tprintf(TEXT("\nCompiling list of all files...\n"));
	for (int i = 0; folders_to_search[i]; i++) {
		findFiles(folders_to_search[i]);
	}
	_tprintf(TEXT("\tFound %d files in %d directories\n"), file_count, dir_count);

	// go through the map and find every case where there is only one file with a particular size
	_tprintf(TEXT("\nElimating all files with no possible identical pair...\n"));
	for (SizeToName::iterator i = m.begin(); i != m.end();) {
		if (i->second.size() == 1) {
			m.erase(i++);
			file_count--;
		} else {
			i++;
		}
	}
	_tprintf(TEXT("\t%d potential duplicates remain\n"), file_count);

	// calculate file hashes for all remaining files
	_tprintf(TEXT("\nCalculating file hashes...\n"));
	int file_count2 = 0;
	for (SizeToName::iterator i = m.begin(); i != m.end(); i++) {
		for (FileList::iterator j = i->second.begin(); j != i->second.end(); j++) {
			j->second = md5hash(j->first, i->first);
			if ((++file_count2) % 1000 == 0) {
				_tprintf(TEXT("\tCalculated %d hashes so far (%d%%)...\n"), file_count2, file_count2 * 100 / file_count);
			}
		}
	}
	_tprintf(TEXT("\tDone calculating file hashes\n"));

	_tprintf(TEXT("\nLooking for duplicate files...\n"));
	FILE *f = _tfopen(TEXT("duplicates.txt"), TEXT("w"));

	// report empty files
	FileList empty_files = m[0];
	SizeToName::iterator iter = m.find(0);
	if (iter != m.end()) { m.erase(iter); }
	if (empty_files.size() > 0) {
		_ftprintf(f, TEXT("Empty Files:\n"));
		for (FileList::iterator j = empty_files.begin(); j != empty_files.end(); j++) {
			_ftprintf(f, TEXT("%s\n"), j->first.c_str());
		}
		_ftprintf(f, TEXT("\n"));
	}

	int last_hundred = 0, files_discarded = 0, files_found = 0;
	LONGLONG size = 0;
	for (SizeToName::iterator i = m.begin(); i != m.end(); i++) {
		FileList v = i->second;
		vector<tstring> temp;
		while (v.size() > 1) {
			tstring cur = v.back().first;
			md5_value cur_hash = v.back().second;
			v.pop_back();
			temp.push_back(cur);
			// go through each file of the same size looking for identical hashes, then check actual identical-ness
			for (FileList::iterator j = v.begin(); j != v.end();) {
				if (!HASH_EQ(cur_hash, j->second)) {
					j++;
					continue;
				}
				int retval = areIdenticalFiles(cur, j->first);
				if (retval == IDENTICAL) {
					temp.push_back(j->first);
					j = v.erase(j);
				} else if (retval == DIFFERENT) {
					j++;
				} else {
					if (retval == FILE_ERR_2 || retval == FILE_ERR_12) {
						j = v.erase(j);
					}
					if (retval == FILE_ERR_1 || retval == FILE_ERR_12) {
						break;
					}
				}
			}
			// found duplicates!
			if (temp.size() > 1) {
				files_found += temp.size();
				if (files_found / 100 > last_hundred) {
					last_hundred = files_found / 100;
					_tprintf(TEXT("\tFound %d00 duplicate files and discared %d files so far (%d%%)...\n"), last_hundred, files_discarded, (files_discarded+last_hundred*100)*100 / file_count);
				}
				sort(temp.begin(), temp.end());
				// save the list of duplicates found
				_ftprintf(f, TEXT("File Size: %s\n"), fileSize(i->first));
				for (vector<tstring>::iterator j = temp.begin(); j != temp.end(); j++) {
					_ftprintf(f, TEXT("%s\n"), j->c_str());
				}
				_ftprintf(f, TEXT("\n"));
				size += i->first * (temp.size() - 1);
			} else {
				files_discarded++;
			}
			temp.clear();
		}
		if (v.size() == 1) {
			files_discarded++;
		}
	}
	_tprintf(TEXT("\tFound %d duplicate files wasting a total of %s bytes\n"), files_found, fileSize(size));

	fclose(f);
	return 0;
}

